;;;  -*- mode: LISP; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Hedderik van Rijn
;;; Copyright   : (c) 2008 Hedderik van Rijn
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; Address     : Department of Artificial Intelligence
;;;             : University of Groningen
;;;             : D.H.van.Rijn@rug.nl
;;; 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : lexical.lisp
;;; Version     : 0.0
;;; 
;;; Description : Implementation of a simple lexicon module.
;;; 
;;; Bugs        : 
;;;
;;; To do       : 
;;; 
;;; ----- History -----
;;;
;;; 2008.03.03 Hedderik
;;;             : Initial creation.
;;; 2008.04.24 Hedderik
;;;             : Added +lexical> isa clear command.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; General Docs:
;;; 
;;; See lexical-documentation.txt
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Public API:
;;;
;;; User Functions:
;;;
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Design Choices:
;;; 
;;;  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; The code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


#+:packaged-actr (in-package :act-r)
#+(and :clean-actr (not :packaged-actr) :ALLEGRO-IDE) (in-package :cg-user)
#-(or (not :clean-actr) :packaged-actr :ALLEGRO-IDE) (in-package :cl-user)

(require-compiled "GOAL-STYLE-MODULE" "ACT-R6:support;goal-style-module")

(defun determine-latency-from-frequency (n)
  (latency (+ (baselevel-optimized n .5 1500000) (act-r-noise .3)) .3))

(defun baselevel-optimized (n d age) 
  (- (log (/ n (- 1 d))) (* d  (log age)))
)

(defun latency (A F) 
  (* F (exp (* -1 A))))

;; ---------------------------------------------------------------------------
;; Module to interact with the lexicon
;;
(defstruct lexical-module 
  lexical-mult
  current-event
  busy
  failed)

(defun create-lexical-module (model-name)
  (declare (ignore model-name))
  (make-lexical-module)
  )

;; This function is called after a (reset) command, initializes
;; everything the module requires. 
(defun lexical-reset (instance)
  (setf (lexical-module-busy instance) nil)
  (setf (lexical-module-failed instance) nil)
  (setf (lexical-module-current-event instance) nil)
  ;; Define the chunk type associated with the lexicon
  (chunk-type lexicon-entry concept lemma frequency reference))

;; Function handling all state queries, i.e., ?lexical> in LHS.
(defun lexical-query (instance buffer-name slot value)
  (declare (ignore buffer-name))
  (case slot
    (state
     (case value
       (busy (lexical-module-busy instance))
       (free (not (lexical-module-busy instance)))
       (failure (lexical-module-failed instance))
       (error nil)
       (t (print-warning "Unknown state query ~S to lexical module" value)
          nil)))
    (t (print-warning "Unknown query of the lexical module"))))

(defvar *chunk-spec* nil)

(defun lexical-request (instance buffer-name chunk-spec)
  (declare (ignore buffer-name))
  (if (and (lexical-module-busy instance)
	   (not (eq 'clear (chunk-spec-chunk-type chunk-spec))))
      ;; Module busy, and this is not a clear request.
      (model-warning "Cannot handle request when busy")
      (progn 
	(setf (lexical-module-failed instance) nil)
	(start-lexical-retrieval instance chunk-spec))))
                                              
(defun start-lexical-retrieval (lexicon request) 
  ;; Clear the "currently in progress" retrieval
  (when (lexical-module-current-event lexicon)
    (delete-event (lexical-module-current-event lexicon))
    (setf (lexical-module-busy lexicon) nil)
    (setf (lexical-module-current-event lexicon) nil))

  ;; When the chunk definition is "clear", we're all set now. Otherwise, schedule a retrieval of a matching chunk.
  (when (not (eq 'clear (chunk-spec-chunk-type request)))
    ;; get the list of all chunks which match the request 
    (let ((chunks (find-matching-chunks request))) 
      (setf (lexical-module-busy lexicon) 
	    (if chunks 
		(progn 
		  ;; Just place the first matching chunk in the buffer  
		  (let ((chunk (car chunks)))
		    (setf (lexical-module-current-event lexicon)
			  (schedule-event-relative (max 0.25 (determine-latency-from-frequency (chunk-slot-value-fct chunk 'frequency))) #'retrieved-lexical-chunk 
						   :module 'lexical 
						   :params (list lexicon chunk)))))
		(setf (lexical-module-current-event lexicon)
		      (schedule-event-relative 1 #'lexical-retrieval-failure 
					       :module 'lexical 
					       :destination 'lexical)))))))
 
;;; Retrieved-chunk 
;;; 
;;; Called as an event when a chunk has been retrieved and is ready to be placed 
;;; into the buffer. 
;;; 
;;; The parameters are an instance of the module and the name of the chunk to put 
;;; into the buffer. 
 
(defun retrieved-lexical-chunk (lexicon chunk) 
   
  ;; Clear the busy flag 
   
  (free-lexical-module lexicon)
  
  ;; Schedule an event to put the chunk into the buffer right now instead of 
  ;; placing it there directly to comply with the guideline that buffer changes 
  ;; should be scheduled. 

  (schedule-set-buffer-chunk 'lexical chunk 0  
                             :module 'lexical  
                             :priority :max)) 

(defun lexical-retrieval-failure (lexicon)
  (free-lexical-module lexicon)
  (setf (lexical-module-failed lexicon) t))

(defun lexical-retrieval-reset-failure (lexicon)
  (free-lexical-module lexicon)
  (setf (lexical-module-failed lexicon) nil))

(defun free-lexical-module (lexicon)
  (setf (lexical-module-busy lexicon) nil)
  (setf (lexical-module-current-event lexicon) nil))

(defun lexical-params (tmp param)
  (cond ((consp param)
         (case (car param)
           (:lexical-mult (setf (lexical-module-lexical-mult tmp) (cdr param)))))
        (t
         (case param
           (:lexical-mult (lexical-module-lexical-mult tmp))))))         


;;; Actually define the module now

(define-module-fct 'lexical '(lexical)
  (list 
        (define-parameter :lexical-mult :valid-test #'posnum :default-value 1
          :warning "a number" :documentation "Lexical multiplier")

        )
  :version "0.1"
  :documentation "The lexical module is used to simulate retrieval of lemmas from the mental lexicon."
  :creation #'create-lexical-module
  :query #'lexical-query
  :request #'lexical-request
  :params #'lexical-params
  :reset #'lexical-reset)



#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#
